/* DBEDITOR.HPP */
/* Copyright Dave Curtis 1994 */
/* dcurtis@hgmp.mrc.ac.uk */
/* no warranty or liability of any kind is accepted, expressed or implied */

/* DBEDITOR.HPP */
/* Copyright Dave Curtis 1991 */
#ifndef DBEDITORHPP
#define DBEDITORHPP 1
#include "dcwindow.hpp"

class dbeditor;

#ifndef using_wx
class db_string_editor :public dcstring_editor
{
int rv;
public:
db_string_editor(int wl=79, int insmode=0, PFC hf=0,int natt=0x07,int hatt=0x70,int ml=241);
int command_source(void);
int edit(int arow,int acol,char *s,int fillin=1,int sc=0,int dc=0,
       int bbc=-1,int bkc=-1,int off=-1);
};
#endif

class edit_field
{
friend class db_editor;
protected:
#ifndef using_wx
static char buffer[241];
#else
wxItem *wx_editor;
int next_x_pixel;
void gen_init(wxPanel *p);
#endif
int row,col,len,winlen;
char *prompt;
void *target;
virtual char *is_error();
virtual void encode();
virtual void decode();
#ifdef using_wx
virtual void init(wxPanel *p);
#endif
public:
int get_len() { return len; } // handy to know if has a target or is just a prompt
edit_field(int r,int c,char *p,char *t,int l,int rl=-1);
virtual ~edit_field() {;}
};

class strdup_field : public edit_field
{
public:
void encode();
void decode();
strdup_field(int r,int c, char *p,char **t,int l,int reallen=-1);
~strdup_field() {;}
};

class dcstring;
class dcstring_field : public edit_field
{
public:
void encode();
void decode();
dcstring_field(int r,int c, char *p,dcstring *t,int l,int reallen=-1);
~dcstring_field() {;}
};

class char_field : public edit_field
{
public:
void encode();
void decode();
char_field(int r,int c, char *p,char *t);
~char_field() {;}
};

class yn_field : public char_field
{
public:
void encode();
void decode();
#if 0
#ifdef using_wx
void init(wxPanel *p); // could have this e.g. for check box
#endif
#endif
char *is_error();
yn_field(int r,int c, char *p,char *t);
~yn_field() {;}
};

class chbox_field : public char_field
{
protected:
wxCheckBox *Box;
public:
void encode();
void decode();
void init(wxPanel *p);
char *is_error();
chbox_field(int r,int c, char *p,char *t);
~chbox_field() { if (Box!=NULL) delete Box; }
};

class double_field : public edit_field
{
double lolim, hilim;
public:
void encode();
void decode();
char *is_error();
double_field(int r,int c, char *p,double *t,int l,double lol=0,double hil=0,int rl=-1);
~double_field() {;}
};

class short_field : public edit_field
{
int lolim, hilim;
public:
void encode();
void decode();
char *is_error();
short_field(int r,int c, char *p,short *t,int l,int lol=0,int hil=0,int rl=-1);
~short_field() {;}
};

class int_field : public edit_field
{
int lolim, hilim;
public:
void encode();
void decode();
char *is_error();
int_field(int r,int c, char *p,int *t,int l,int lol=0,int hil=0,int rl=-1);
~int_field() {;}
};

typedef edit_field *pedit_field;
declare (dczGDList,edit_field);

#ifndef using_wx
class db_editor : public dcwindow
{
void (*helpfunc)(char *,...);
pedit_field kept_pe;
protected:
db_string_editor le;
dcwindow exp_win;
virtual int edit(pedit_field pe,int r,int c);
unsigned char page,npages;
char title_str[100];
public:
dczGDList(edit_field) fieldlist;
db_editor(dcwindow_set *ws,char *t,int r1,int c1,int r2,int c2,int np=1,int att=DISP_NORMAL,int fr=0,int fratt=-1,void (*hf)(char *form,...)=0);
void explain(char *s);
int edit(int st=-1,int allow_next_or_prev=0);
// NB!! With wx, p is a page number, with DOS st is a field number
show_fields();
do_editing(void *thing_to_edit,int allow_next_or_prev=0);
virtual int check()=0;
virtual int copy_in(void *source)=0;
virtual int copy_out(void *dest)=0;
};

#else
#define MAXDBPAGES 20
class modal_panel : public wxDialogBox {
public:
modal_panel(wxFrame *Parent, char *Title,
                         int x, int y, int width, int height);
};

class db_editor
{
void (*helpfunc)(char *,...);
protected:
int page,npages;
wxFrame **parentptr;
modal_panel *panel[MAXDBPAGES];
char inited,active;
int r1,r2,c1,c2;
init(int allow_next_or_prev);
char title_str[100];
public:
dczGDList(edit_field) fieldlist;
db_editor(wxFrame **p,char *t,int r1,int c1,int r2,int c2,int np=1);
virtual ~db_editor();
void explain(char *s);
int edit(int p=-1,int allow_next_or_prev=0);
// NB!! With wx, p is a page number, with DOS it is a field number
show_fields();
do_editing(void *thing_to_edit,int allow_next_or_prev=0);
virtual int check()=0;
virtual int copy_in(void *source)=0;
virtual int copy_out(void *dest)=0;
void on_ok();
void on_cancel();
void on_prev_p();
void on_next_p();
void on_next_rec();
void on_prev_rec();
};

extern void set_scale(wxWindow *w);

#endif

#endif

