#include "dcerror.hpp"
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <time.h>
#ifdef MDB
#include <mem.h>
#endif

#include "dbf.hpp"
#include "dcindex.hpp"
#include "dcutils.hpp"
#include "dbexpr.hpp"

#ifndef max
#define max(x,y) ((x)>(y)?(x):(y))
#define min(x,y) ((x)<(y)?(x):(y))
#endif

#ifdef BYTESWAP
#define SWAP(x,y,t) (t=x,x=y,y=t)
void longswap(DCDWORD *l)
{ char t,*p; p=(char *)l; SWAP(p[0],p[3],t); SWAP(p[1],p[2],t); }
void shortswap(unsigned short *s)
{ char t,*p; p=(char *)s; SWAP(p[0],p[1],t); }
#endif

int db3_header::make_hdc(char *fn)
{
int r=1,c=12;
data_file fo;
if (!fo.open(fn,"w")) return 0;
// fo.fprintf("TOP=0\nLEFT=0\nHIGH=18\nWIDE=78\n");
for (int i=0;i<num_fields;++i)
  {
  db3_field_defn *def;
  def=flist[i];
  fo.fprintf("N%s\n",def->title);
  fo.fprintf("T%c\n",def->field_type);
  fo.fprintf("L%d\n",def->length);
  fo.fprintf("C%d\n",c);
  fo.fprintf("R%d\n",r);
  if (def->length>20) fo.fprintf("V20\n");
  if (c==12) c=52;
  else
    { c=12; ++r; }
  }
return 1;
}

db3_field_defn::db3_field_defn(db3_field_defn &old)
{
memmove(this,&old,sizeof(db3_field_defn));
}

int db3_header::rec_is_deleted(long r)
{
if (!goto_rec(r?r:this_rec)) return 0;
int c=file->fgetc();
ungetc(c,get_fp(*file));
return c=='*';
}

int db3_record_buffer::make_index(char *fn,char *s)
{
dc_index idx;
if (!idx.make_new(fn)) return 0;
db_express e(this);
if (!e.parse(s)) return 0;
for (long r=1L;r<=header->get_num_recs();++r)
  {
  header->goto_rec(r);
  header->record_load(*this);
  if (!is_deleted())
    {
    dcexpr_val *rv=e.eval();
    if (dcerror.stat() || !idx.add((char*)(*rv),r))
      {
      delete rv;
      idx.close();
      return 0;
      }
    delete rv;
    }
  }
idx.close();
return 1;
}

field_array::field_array(int dim) : gdcarray(db3_field_defn)(dim) {;}

db3_field_defn::db3_field_defn() {;}

extern "C" {
int db3_field_defn_compare(db3_field_defn **l1,db3_field_defn**l2)
{ return 0; } } // never used, but some compilers want it

int db3_header::which_field(char *f)
{
return flist.which(f);
}

int db3_record_buffer::add_string(int fnum,char *data)
{
char *ptr;
while ((ptr=strchr(data,'\n'))!=0) *ptr=' ';
db3_field_defn* def=header->flist[fnum];
strncpy(buff+def->offset,data,min(strlen(data),def->real_length));
for (int i=strlen(data);i<def->real_length;++i) buff[def->offset+i]=' ';
return 1;
}

int db3_record_buffer::add_string(char *fname,char *data)
{
int fnum=header->flist.which(fname);
return fnum==-1?0:add_string(fnum,data);
}

db3_field_value::db3_field_value(char *d,unsigned short l,char t)
{ data=d; len=l; type=t; }


db3_field_value db3_record_buffer::operator[](int ind)
  {
  db3_field_defn* def=header->flist[ind];
  return db3_field_value(buff+def->offset,def->real_length,def->field_type);
  }

db3_field_value db3_record_buffer::operator()(char *title)
  {
  int w;
  w=header->flist.which(title);
  if (w==-1)
    return db3_field_value(buff,0,'C');
  else
    {
    db3_field_defn* def=header->flist[w];
    return db3_field_value(buff+def->offset,def->real_length,def->field_type);
    }
  }

db3_record_buffer::db3_record_buffer(db3_header *h)
{ header=h; buff=new char[h->data_len-1]; clear(); }

void db3_record_buffer::clear()
 { deleted=' '; memset(buff,' ',header->data_len-1); } // '*' to delete

char *db3_field_defn::label() { return title; }

db3_field_defn::db3_field_defn(char *t,char f,unsigned short l,DCDWORD o,
   unsigned char d,int *z)
{
sprintf(title,"%.10s",t);
title[10]='\0';
strupr(title);
field_type=f;
length=real_length=l; decimal=d; offset=o;
if (z) memcpy(zeros,z,7*sizeof(short));
else memset(zeros,0,7*sizeof(short));
}

int db3_header::record_save(db3_record_buffer &rb,unsigned long r)
{
if (!goto_rec(r?r:this_rec)) return 0;
if (!fwrite((char*)&rb.deleted,1,1,get_fp(*file))) return 0;
if (!fwrite((char*)rb.buff,data_len-1,1,get_fp(*file))) return 0;
return 1;
}

int db3_header::record_load(db3_record_buffer &rb,unsigned long r)
{
if (!goto_rec(r?r:this_rec)) return 0;
if (!fread((char*)&rb.deleted,1,1,get_fp(*file))) return 0;
if (!fread((char*)rb.buff,data_len-1,1,get_fp(*file))) return 0;
return 1;
}

long db3_memo_file::add_memo(char *s)
{
int l;
long old_off=mem_offset;
fseek(fp,mem_offset*512,SEEK_SET);
fwrite(s,l=strlen(s),1,fp);
fwrite("\x1a\x1a",2,1,fp);
l+=2;
mem_offset+=(l-1)/512+1;
fwrite(zero_buffer,(10240-l)%512,1,fp);
return old_off; // to go into record buffer
}

char *db3_memo_file::read_memo(long r)
{
unsigned len;
int c;
char *buff;
if (fseek(fp,r*512,SEEK_SET)) return 0;
for (len=0;(c=fgetc(fp))!=EOF && c!=0x1a;++len) ;
if ((buff=new char[len+1])==0) return 0;
if (fseek(fp,r*512,SEEK_SET)) { delete buff; return 0; }
for (len=0;(c=fgetc(fp))!=EOF && c!=0x1a;++len) buff[len]=c;
buff[len]='\0';
return buff;
}

db3_memo_file::db3_memo_file(FILE *f)
{
fp=f;
fseek(fp,0L,SEEK_END);
if (ftell(fp)==0l)
  {
  memset(zero_buffer,' ',sizeof(zero_buffer));
  fwrite(zero_buffer,512,1,fp);
  mem_offset=1l;
  }
else
  {
  fseek(fp,0l,SEEK_SET);
  fread((char*)&mem_offset,sizeof(long),1,fp);
  fseek(fp,mem_offset*512,SEEK_SET);
  }
}

db3_header::db3_header(data_file *f,unsigned char v,db3_memo_file *mf)
{
ver=v;
file=f;
memofile=mf;
num_recs=0;
num_fields=0;
data_off=32+1; // for end of header marker
data_len=1; // for delete field
// memset(blank,' ',20);
memset(blank,0,20);
end_header=0xd;
set_date();
}

void db3_header::set_date()
{
time_t t;
struct tm *tmp;
time(&t);
tmp=localtime(&t);
set_date(tmp->tm_mday,tmp->tm_mon+1,tmp->tm_year);
}

int db3_field_defn::input(data_file *f)
{
if (fread((char*)&title,sizeof(title),1,get_fp(*f))!=1)
  return 0;
if (fread((char*)&field_type,sizeof(field_type),1,get_fp(*f))!=1)
  return 0;
if (fread((char*)&offset,sizeof(offset),1,get_fp(*f))!=1)
  return 0;
if (fread((char*)&length,sizeof(length),1,get_fp(*f))!=1)
  return 0;
if (fread((char*)&decimal,sizeof(decimal),1,get_fp(*f))!=1)
  return 0;
if (fread((char*)&zeros,sizeof(zeros),1,get_fp(*f))!=1)
  return 0;

if (title[0]==0x0d) return 0;
for (int i=0;i<11;++i)
  if (title[i]=='\0') break;
  else if (title[i]==' ') { title[i]='\0'; break; }
strupr(title);
#ifdef BYTESWAP
longswap(&offset);
#endif
real_length=length;
return 1;
}

int db3_field_defn::output(data_file *f)
{
int rv;
for (int i=0;i<11;++i)
  if (title[i]=='\0' || title[i]==' ')
    { while (i++<11) fputc('\0',get_fp(*f)); break; }
  else fputc(title[i],get_fp(*f));
#ifdef BYTESWAP
longswap(&offset);
#endif
rv=1;
// rv=rv && fwrite((char*)&title,sizeof(title),1,get_fp(*f))==1;
rv=rv && fwrite((char*)&field_type,sizeof(field_type),1,get_fp(*f))==1;
rv=rv && fwrite((char*)&offset,sizeof(offset),1,get_fp(*f))==1;
rv=rv && fwrite((char*)&length,sizeof(length),1,get_fp(*f))==1;
rv=rv && fwrite((char*)&decimal,sizeof(decimal),1,get_fp(*f))==1;
rv=rv && fwrite((char*)&zeros,sizeof(zeros),1,get_fp(*f))==1;
#ifdef BYTESWAP
longswap(&offset);
#endif
return rv;
}

void db3_header::save_num_recs(DCDWORD nr)
{
file->fseek(4l,SEEK_SET);
#ifdef BYTESWAP
longswap(&nr);
#endif
fwrite((char*)&nr,4,1,get_fp(*file));
}

int db3_header::load()
{
db3_field_defn temp;
file->fseek(0l,SEEK_SET);
if (fread((char*)&ver,sizeof(ver),1,get_fp(*file))!=1)
 { dcerror(1,"Error reading header"); return 0; }
if (fread((char*)&date,sizeof(date),1,get_fp(*file))!=1)
 { dcerror(1,"Error reading header"); return 0; }
if (fread((char*)&num_recs,sizeof(num_recs),1,get_fp(*file))!=1)
 { dcerror(1,"Error reading header"); return 0; }
if (fread((char*)&data_off,sizeof(data_off),1,get_fp(*file))!=1)
 { dcerror(1,"Error reading header"); return 0; }
if (fread((char*)&data_len,sizeof(data_len),1,get_fp(*file))!=1)
 { dcerror(1,"Error reading header"); return 0; }
if (fread((char*)&blank,sizeof(blank),1,get_fp(*file))!=1)
 { dcerror(1,"Error reading header"); return 0; }
#ifdef BYTESWAP
longswap(&num_recs);
shortswap(&data_off);
shortswap(&data_len);
#endif
flist.clear();
num_fields=0;
// unsigned long old_off=0,old_len=0;
while (temp.input(file) &&  file->ftell()<data_off)
  {
  if (num_fields && !temp.offset) // ??old versions of PC-File
     temp.offset=flist[num_fields-1]->offset+flist[num_fields-1]->length;
  flist.add_member(new db3_field_defn(temp));
  ++num_fields;
  }
flist.add_member(new
  db3_field_defn("ALL_FIELDS",'C',data_len-1,0));
if (file->fseek(data_off,SEEK_SET)!=0)
 { dcerror(1,"Seek error after reading header"); return 0; }
this_rec=1L;
return 1;
}

int db3_header::goto_rec(unsigned long r)
{
if (r<1) { this_rec=0L; return 0; }
// may be deliberately set to mean we are nowhere
if (file->fseek(data_off+(r-1L)*data_len,SEEK_SET))
   return 0; // always fseek() in case changing from read to write
this_rec=r;
return 1;
}

int db3_header::goto_first(dc_index *i,int skip_deleted)
{
if (!i)
  {
  if (!skip_deleted) return goto_rec(1L);
  else
    {
    for (long r=1L;r<=num_recs;++r)
      if (!rec_is_deleted(r)) return 1;  // does a goto_rec(r)
    return goto_rec(0L); // all records deleted, we're nowhere
    }
  }
else
 {
 long r=i->get_first();
 return goto_rec(r); // returns 0 if r==0L
 }
}

int db3_header::goto_last(dc_index *i,int skip_deleted)
{
if (!i)
  {
  if (!skip_deleted) return goto_rec(num_recs);
  else
    {
    for (long r=num_recs;r>=1L;--r)
      if (!rec_is_deleted(r)) return 1;  // does a goto_rec(r)
    return goto_rec(0L); // all records deleted
    }
  }
else
 {
 long r=i->get_last();
 return goto_rec(r);
 }
}

int db3_header::goto_near_match(char *m,dc_index *i)
{
if (!i)
  {
  dcerror(1,"No index in use for goto_near_match(%s)",m);
  return 0;
  }
else
 {
 long r=i->near_find(m);
 return goto_rec(r);
 }
}

int db3_header::goto_exact_match(char *m,dc_index *i)
{
if (!i)
  { dcerror(1,"No index in use for goto_exact_match(%s)",m); return 0; }
else
 {
 long r=i->exact_find(m);
 if (!r)
   dcerror(1,"No match for goto_exact_match(%s)",m);
 return goto_rec(r);
 }
}

int db3_header::goto_prev(dc_index *i,int skip_deleted)
{
if (!i)
  {
  if (this_rec==1L) return 0; // already at start
  else if (!skip_deleted) return goto_rec(this_rec-1L);
  else
    {
    long r,curr_rec=this_rec;
    for (r=this_rec-1L;r>=1L;--r)
      if (!rec_is_deleted(r)) return 1;
    goto_rec(curr_rec);
    return 0;
    }
  }
else
 {
 long r=i->get_prev();
 return goto_rec(r);
 }
}

int db3_header::goto_next(dc_index *i,int skip_deleted)
{
// NB this compares with num_recs, nothing else does, will fail if
// have added records and haven't increased num_recs
if (!i)
  {
  if (this_rec==num_recs) return 0; // already at end
  else if (!skip_deleted) return goto_rec(this_rec+1L);
  else
    {
    long r,curr_rec=this_rec;
    for (r=this_rec+1L;r<=num_recs;++r)
      if (!rec_is_deleted(r)) return 1;
    goto_rec(curr_rec); // none found
    return 0;
    }
  }
else
 {
 long r=i->get_next();
 return goto_rec(r);
 }
}

int db3_header::save()
{
int rv=1;
#ifdef BYTESWAP
longswap(&num_recs);
shortswap(&data_off);
shortswap(&data_len);
#endif
file->fseek(0l,SEEK_SET);
rv=1;
rv=rv && fwrite((char*)&ver,sizeof(ver),1,get_fp(*file))==1;
rv=rv && fwrite((char*)&date,sizeof(date),1,get_fp(*file))==1;
rv=rv && fwrite((char*)&num_recs,sizeof(num_recs),1,get_fp(*file))==1;
rv=rv && fwrite((char*)&data_off,sizeof(data_off),1,get_fp(*file))==1;
rv=rv && fwrite((char*)&data_len,sizeof(data_len),1,get_fp(*file))==1;
rv=rv && fwrite((char*)&blank,sizeof(blank),1,get_fp(*file))==1;
if (rv) for (int i=0;i<num_fields;++i)
  if (!flist[i]->output(file)) { rv=0; break; }
#ifdef BYTESWAP
longswap(&num_recs);
shortswap(&data_off);
shortswap(&data_len);
#endif
if (rv && fwrite((char*)&end_header,1,1,get_fp(*file))!=1) rv=0;
if (!rv)
 { dcerror(1,"Error saving header"); return 0; }
else this_rec=1L;
return rv;
}

int db3_header::add_field(char *t,char f,unsigned char l)
{
flist.add_member(new db3_field_defn(t,f,l,data_len-1));
data_len+=l;
data_off+=32;
return ++num_fields;
}

char db3_memo_file::zero_buffer[512];


