/* WXDBEDIT.CXX */
/* Copyright Dave Curtis 1994 */
/* dcurtis@hgmp.mrc.ac.uk */
/* no warranty or liability of any kind is accepted, expressed or implied */

/* DBEDITOR.CPP */

/* Copyright Dave Curtis 1991 */
// #include "allwx.hpp"
#include "dcerror.hpp"
#include <stdlib.h>
#include <limits.h>
#include "dcstr.hpp"
#include <ctype.h>
#include <math.h>
#include "dbeditor.hpp"
#include "assert.h"

#define DEFAULT_HSPACING 2

static char error_buff[60];

class db_editor_button: public wxButton
{
 public:
  db_editor *db;
  db_editor_button(db_editor *d,wxPanel *panel, wxFunction func, char *label,
               int x = -1, int y = -1, int width = -1, int height = -1);
};

db_editor_button::db_editor_button(db_editor *d,wxPanel *panel, wxFunction func, char *label,
             int x, int y, int width, int height):
  wxButton(panel, func, label, x, y, width, height)
{ db=d; }

void db_editor_ok(db_editor_button& button, wxEvent& event);
void db_editor_cancel(db_editor_button& button, wxEvent& event);
void db_editor_next_p(db_editor_button& button, wxEvent& event);
void db_editor_prev_p(db_editor_button& button, wxEvent& event);
void db_editor_next_rec(db_editor_button& button, wxEvent& event);
void db_editor_prev_rec(db_editor_button& button, wxEvent& event);

void db_editor_ok(db_editor_button& button, wxEvent& event)
  { button.db->on_ok(); }
void db_editor_cancel(db_editor_button& button, wxEvent& event)
  { button.db->on_cancel(); }
void db_editor_next_p(db_editor_button& button, wxEvent& event)
  { button.db->on_next_p(); }
void db_editor_prev_p(db_editor_button& button, wxEvent& event)
  { button.db->on_prev_p(); }
void db_editor_next_rec(db_editor_button& button, wxEvent& event)
  { button.db->on_next_rec(); }
void db_editor_prev_rec(db_editor_button& button, wxEvent& event)
  { button.db->on_prev_rec(); }

int db_editor::do_editing(void *source,int allow_next_or_prev)
{
copy_in(source);
int c,ret;
do
  {
  if ((ret=edit(-1,allow_next_or_prev))==0) return 0;
  } while ((c=check())!=0);
copy_out(source);
return ret;
}

db_editor::db_editor(wxFrame **p,char *t,int rr1,int cc1,int rr2,int cc2,int np)
{
npages=np;
for (page=0;page<npages;++page)
  panel[page]=0;
page=0;
inited=0;
active=0;
strcpy(title_str,t);
r1=rr1; c1=cc1; r2=rr2; c2=cc2;
parentptr=p?p:(wxFrame**)&mainframeptr;
}

int db_editor::init(int allow_next_or_prev)
{
dczGDList(edit_field) temp_fl;
int f;
// pathetic sort
while (fieldlist.size()) temp_fl<<*fieldlist;
while (temp_fl.size())
  {
  int minr=9999,minc=9999,minf;
  temp_fl.start();
  for (f=0;f<temp_fl.size();++f,++temp_fl)
    if (temp_fl()->row<minr || (temp_fl()->row==minr && temp_fl()->col<minc))
      { minf=f; minr=temp_fl()->row; minc=temp_fl()->col; }
  temp_fl.start();
  for (f=0;f<minf;++f) ++temp_fl;
  fieldlist.end();
  fieldlist<<*temp_fl;
  }
for (int p=0;p<npages;++p)
  {
  char buff[100];
  if (npages==1) strcpy(buff,title_str);
  else sprintf(buff,"%s (page %d of %d)",title_str,p+1,npages);
#ifdef wx_msw
//  panel[p]=new modal_panel(*parentptr,buff,c1*xscale,r1*yscale,(c2-c1)*xscale*1.2,(r2-r1+4)*yscale*1.3);
  panel[p]=new modal_panel(*parentptr,buff,c1*xscale,r1*yscale,(c2-c1)*xscale*1.2,(r2-r1+6)*yscale*1.05);
#else
#ifdef wx_motif
  panel[p]=new modal_panel(*parentptr,buff,int(c1*xscale),int(r1*yscale),int((c2-c1)*xscale*1.1),int((r2-r1+3)*yscale*1.5));
#else
  panel[p]=new modal_panel(*parentptr,buff,int(c1*xscale),int(r1*yscale),int((c2-c1)*xscale*1.1),int((r2-r1+3)*yscale*1.3));
#endif
#endif
  assert(panel[p]!=NULL);
  panel[p]->SetVerticalSpacing((int)(yscale*0.0));
  wxFont *smallFont;
  smallFont=wxTheFontList->FindOrCreateFont(11,wxTELETYPE,wxNORMAL,wxNORMAL);
  assert(smallFont!=NULL);
  panel[p]->SetLabelFont(smallFont);
  panel[p]->SetButtonFont(smallFont);
#ifdef wx_motif
  panel[p]->SetLabelPosition(wxHORIZONTAL);
  panel[p]->SetHorizontalSpacing(DEFAULT_HSPACING);
#endif
  // this here to start buttons right at the top
  assert(new db_editor_button(this,panel[p],(wxFunction)db_editor_ok,"OK"));
  assert(new db_editor_button(this,panel[p],(wxFunction)db_editor_cancel,"Cancel"));
  if (allow_next_or_prev)
    {
    assert(new db_editor_button(this,panel[p],(wxFunction)db_editor_prev_rec,"OK+Previous"));
    assert(new db_editor_button(this,panel[p],(wxFunction)db_editor_next_rec,"OK+Next"));
    }
  if (p<npages-1)
    assert(new db_editor_button(this,panel[p],(wxFunction)db_editor_next_p,"Page down"));
  if (p>0)
    assert(new db_editor_button(this,panel[p],(wxFunction)db_editor_prev_p,"Page up"));
  //panel[p]->NewLine();
  int cur_x;
  panel[p]->SetHorizontalSpacing(0);
  cur_x=0;
  int do_new_line=1,new_line_size=0;
  int left_margin=PANEL_HSPACING;  // may be different under X or in the future

//   new wxMessage(panel[p],""); has no effect
  int thisrow=p*(r2-r1-2);
  fieldlist.start();
  while (fieldlist()->row<thisrow) ++fieldlist;
  while (fieldlist()->row<(p+1)*(r2-r1-2))
    {
    while (thisrow<fieldlist()->row)
      {
      ++thisrow;
      if (do_new_line) new_line_size+=int(yscale);
      else do_new_line=1;
      //panel[p]->SetVerticalSpacing(panel_VSPACING);
      //panel[p]->NewLine();
      //new wxMessage(panel[p],".");
      //panel[p]->SetVerticalSpacing(0);
      assert(new wxMessage(panel[p],""));
      cur_x=0;
      }
    if (do_new_line)
      {
      panel[p]->SetVerticalSpacing(new_line_size);
      panel[p]->SetHorizontalSpacing(0);
      panel[p]->NewLine(); // produces x position of PANEL_HSPACING (at present!)
      new_line_size=0;
//      new_line_size=int(yscale)-panel[p]->GetCharHeight();
      do_new_line=0;
      }
    if (cur_x==0 && DEFAULT_HSPACING<fieldlist()->col*xscale*1.05+left_margin)
      {
      wxWindow *w;
      int posx,x,posy,y;
      w=new wxMessage(panel[p]," ");
      assert(w!=NULL);
      w->GetPosition(&posx,&posy);
      w->GetSize(&x,&y);
      cur_x=x+posx;
      }
    if (cur_x+DEFAULT_HSPACING<fieldlist()->col*xscale*1.05+left_margin)
// stretch out by 1.05 to avoid items crowding each other out
#if 0
       panel[p]->Tab(fieldlist()->col*xscale-cur_x-DEFAULT_HSPACING);
#else
       panel[p]->SetHorizontalSpacing(int(fieldlist()->col*xscale*1.05+left_margin-cur_x));
#endif
    fieldlist()->init(panel[p]);
    panel[p]->SetHorizontalSpacing(DEFAULT_HSPACING); // messed up by Tab()
    cur_x=fieldlist()->next_x_pixel;
//    panel[p]->GetCursor(&cur_x,&cur_y);
    if (fieldlist.atend()) break;
    else ++fieldlist;
    }
  }
inited=1;
return 1;
}

db_editor::~db_editor()
{
while (fieldlist.size()) delete *fieldlist;
for (page=0;page<npages;++page)
  if (panel[page]) delete panel[page];
}

#if defined DEBUG && defined wx_msw
#include <toolhelp.h>
#endif
int db_editor::edit(int p,int allow_next_or_prev)
{
if (xscale==0.0) set_scale((wxWindow *)(*parentptr));
if (!inited) init(allow_next_or_prev);
active=-1;
if (p>=0) page=p; // otherwise use last value
if (page>npages-1) page=npages-1;
#if defined DEBUG && defined wx_msw
FILE *res=fopen("resusage.tmp","a");
SYSHEAPINFO info;
info.dwSize=sizeof(info);
#endif
int f;
for (f=0,fieldlist.start();f<fieldlist.size();++f,++fieldlist)
        {
        if (fieldlist()->len==0) continue;
        fieldlist()->decode();
        }
while (active==-1)
  {
  int oldpage=page;
#if 0
  int f;
  for (f=0,fieldlist.start();f<fieldlist.size();++f,++fieldlist)
        {
        if (fieldlist()->len==0) continue;
        if (fieldlist()->row<page*(r2-r1-2) ||
            fieldlist()->row>=(page+1)*(r2-r1-2)) continue;
        fieldlist()->decode();
        }
#endif
  while (oldpage==page && active==-1)
    {
    panel[page]->Show(TRUE);
    if (active>0)
      {
      char *err=0;
      for (f=0,fieldlist.start();f<fieldlist.size();++f,++fieldlist)
        {
        if (fieldlist()->len==0) continue;
#if 0
        if (fieldlist()->row<page*(r2-r1-2) ||
            fieldlist()->row>=(page+1)*(r2-r1-2)) continue;
        else
#endif
        if ((err=fieldlist()->is_error())!=0)
          { explain(err); active=-1; break; }
        }
      if (err==0) for (f=0,fieldlist.start();f<fieldlist.size();++f,++fieldlist)
        if (fieldlist()->len!=0) fieldlist()->encode();
      }
    }
  }
#if defined DEBUG && defined wx_msw
fclose(res);
#endif
return active;
}

void db_editor::on_cancel()
{
active=0;
panel[page]->Show(FALSE);
}

void db_editor::on_next_p()
{
if (page<npages-1)
  { panel[page]->Show(FALSE); ++page; }
}

void db_editor::on_prev_p()
{
if (page>0)
  { panel[page]->Show(FALSE); --page; }
}

void db_editor::on_ok()
{
if (active) active=1;
panel[page]->Show(FALSE);
}

void db_editor::on_prev_rec()
{
if (active) active=2;
panel[page]->Show(FALSE);
}

void db_editor::on_next_rec()
{
if (active) active=3;
panel[page]->Show(FALSE);
}

void db_editor::explain(char *s)
{ wxMessageBox(s,"Data editor message"); }

edit_field::edit_field(int r,int c, char *p,char *t,int l,int rl)
{ row=r; col=c; winlen=l; len=rl==-1?l:rl; prompt=p; target=t; }

modal_panel::modal_panel(wxFrame *Parent, char *Title,
                         int x, int y, int width, int height) :
wxDialogBox(Parent,Title,TRUE,x,y,width,height)
{ Show(FALSE); }

void edit_field::encode()
{
char *s=((wxText*)wx_editor)->GetValue();
strncpy((char*)target,s,len);
*((char*)target+len)='\0';
}

char_field::char_field(int r,int c, char *p,char *t)
 : edit_field(r,c,p,(char *)t,1)
{ ; }

yn_field::yn_field(int r,int c, char *p,char *t)
 : char_field(r,c,p,t)
{ ; }

#include <assert.h>
void edit_field::gen_init(wxPanel *p)
{
wxWindow *w;
int posx,widx,y;
if (*prompt!='\0')
  {
  w=new wxMessage(p,prompt);
  p->SetHorizontalSpacing(DEFAULT_HSPACING); // messed up by Tab()
  }
if (len!=0)
  {
#ifndef OLD_WAY
  w=wx_editor=new wxText(p,NULL,NULL,"",-1,-1,int((winlen+2)*xscale),-1);
  assert(w);
#else
  w->GetPosition(&posx,&y);
  w->GetSize(&widx,&y);
  if (posx+widx+DEFAULT_HSPACING<(col+strlen(prompt))*xscale*1.05+left_margin)
    p->SetHorizontalSpacing((col+strlen(prompt))*xscale*1.05+left_margin-(posx+widx+DEFAULT_HSPACING));
  w=wx_editor=new wxText(p,NULL,NULL," ",-1,-1,int((winlen+2)*xscale),int(yscale*1.5));
#endif
  }
w->GetPosition(&posx,&y);
w->GetSize(&widx,&y);
next_x_pixel=posx+widx;
}

void edit_field::init(wxPanel *p)
{ gen_init(p); }

void edit_field::decode()
{ ((wxText*)wx_editor)->SetValue((char*)target); }

void yn_field::decode()
{
char buff[2];
buff[0]=(*(char*)target)?'Y':'N'; buff[1]='\0';
((wxText*)wx_editor)->SetValue(buff);
}

void yn_field::encode()
{
char ans=*((wxText*)wx_editor)->GetValue();
(*(char*)target)=isalpha(ans) && 'Y'==(islower(ans)?toupper(ans):ans);
}

void char_field::encode()
{
char *s=((wxText*)wx_editor)->GetValue();
*(char*)target=*s;
}

void char_field::decode()
{
char buff[2];
buff[0]=*(char*)target; buff[1]='\0';
((wxText*)wx_editor)->SetValue(buff);
}

double_field::double_field(int r,int c, char *p,double *t,int l,double lol,double hil,int rl)
 : edit_field(r,c,p,(char *)t,l,rl)
{ if (!lol && !hil) lol=-(hil=HUGE_VAL); lolim=lol; hilim=hil; }

void double_field::decode()
{
char format[20],buff[50];
#if 0
sprintf(format,"%%.%dg",len-5);
sprintf(buff,format,*(double*)target);
#else
sprintf(buff,"%.5g",*(double*)target);
#endif
((wxText*)wx_editor)->SetValue(buff);
}

void double_field::encode()
{
char *s=((wxText*)wx_editor)->GetValue();
*(double*)target=atof(s);
}

int_field::int_field(int r,int c, char *p,int *t,int l,int lol,int hil,int rl)
 : edit_field(r,c,p,(char *)t,l,rl)
{ if (!lol && !hil) lol=-(hil=INT_MAX); lolim=lol; hilim=hil; }

void int_field::decode()
{
char buff[50];
sprintf(buff,"%d",*(int*)target);
((wxText*)wx_editor)->SetValue(buff);
}

void int_field::encode()
{
char *s=((wxText*)wx_editor)->GetValue();
*(int*)target=atoi(s);
}

short_field::short_field(int r,int c, char *p,short *t,int l,int lol,int hil,int rl)
 : edit_field(r,c,p,(char *)t,l,rl)
{ if (!lol && !hil) lol=-(hil=INT_MAX); lolim=lol; hilim=hil; }

void short_field::decode()
{
char buff[50];
sprintf(buff,"%d",*(short*)target);
((wxText*)wx_editor)->SetValue(buff);
}

void short_field::encode()
{
char *s=((wxText*)wx_editor)->GetValue();
*(short*)target=atoi(s);
}

strdup_field::strdup_field(int r,int c, char *p,char **t,int l,int rl)
 : edit_field(r,c,p,(char *)t,l,rl) { ; }

void strdup_field::decode()
{
if (*(char**)target)
  ((wxText*)wx_editor)->SetValue(*(char**)target);
else ((wxText*)wx_editor)->SetValue("");
}

void strdup_field::encode()
{
if (*(char**)target) { free(*(char**)target); *(char**)target=0; }
char *s=((wxText*)wx_editor)->GetValue();
if (strlen(s) && (*(char**)target=strdup(s))==0)
   dcerror(ENOMEM);
}

dcstring_field::dcstring_field(int r,int c, char *p,dcstring *t,int l,int rl)
 : edit_field(r,c,p,(char *)t,l,rl) { ; }

void dcstring_field::decode()
{
if (!((dcstring *)target)->is_null())
  ((wxText*)wx_editor)->SetValue(*(char**)target);
else ((wxText*)wx_editor)->SetValue("");
}

void dcstring_field::encode()
{
((dcstring *)target)->strfree();
char *s=((wxText*)wx_editor)->GetValue();
if (strlen(s) && !((dcstring*)target)->assign(s))
   dcerror(ENOMEM);
}

char *yn_field::is_error()
{
char c=toupper(*((wxText*)wx_editor)->GetValue());
if (c!='Y' && c!='N')
   {
   sprintf(error_buff,"Illegal character '%c' in Y/N field",c);
   return error_buff;
   }
else return 0;
}

char *double_field::is_error()
{
double i=atof(((wxText*)wx_editor)->GetValue());
if (i<lolim || i>hilim)
  {
  sprintf(error_buff,"Value %f is out of range (%f - %f)",
              atof(((wxText*)wx_editor)->GetValue()),lolim,hilim);
  return error_buff;
  }
else return 0;
}

char *int_field::is_error()
{
int i=atoi(((wxText*)wx_editor)->GetValue());
if (i<lolim || i>hilim)
  {
  sprintf(error_buff,"Value %d is out of range (%d - %d)",
              atoi(((wxText*)wx_editor)->GetValue()),lolim,hilim);
  return error_buff;
  }
else return 0;
}

char *short_field::is_error()
{
int i=atoi(((wxText*)wx_editor)->GetValue());
if (i<lolim || i>hilim)
  {
  sprintf(error_buff,"Value %d is out of range (%d - %d)",
              atoi(((wxText*)wx_editor)->GetValue()),lolim,hilim);
  return error_buff;
  }
else return 0;
}

char *edit_field::is_error()
{
return 0;
}

chbox_field::chbox_field(int r,int c, char *p,char *t)
 : char_field(r,c,p,t)
{ Box=NULL; }

void chbox_field::init(wxPanel *p)
{
wxWindow *w;
int posx,widx,y;
if (*prompt!='\0')
  {
  new wxMessage(p,prompt);
  p->SetHorizontalSpacing(DEFAULT_HSPACING); // messed up by Tab()
  }
w=Box=new wxCheckBox(p,NULL,"");
w->GetPosition(&posx,&y);
w->GetSize(&widx,&y);
next_x_pixel=posx+widx;
}

void chbox_field::decode()
{
Box->SetValue(*(char*)target);
}

void chbox_field::encode()
{
(*(char*)target)=Box->GetValue();
}

char *chbox_field::is_error()
{
return 0;
}


