/* WXDCWIN.CXX */ 
/* Copyright Dave Curtis 1994 */ 
/* dcurtis@hgmp.mrc.ac.uk */ 
/* no warranty or liability of any kind is accepted, expressed or implied */ 
 
#include <dcwindow.hpp>
#include <winmenu.hpp>
#include <string.h>
#include <ctype.h>
#ifdef __ZTC__
#include <dos.h>
#endif

#define DEFAULT_MAX_INPUT_LENGTH 60

float xscale=0.0,yscale=0.0;

extern char *wxGetTextWithHistory(char *message, char *caption = "Input text",
                        char *default_value = "", wxFrame *parent = NULL,
                        int x = -1, int y = -1);

dcButton::dcButton(wxDialogBox *dd,wxPanel *panel, wxFunction func, char *label,
             int x, int y, int width, int height):
  wxButton(panel, func, label, x, y, width, height)
{ d=dd; }

int dcDialogButtonPressed=wxCANCEL;

void dcDialogOkButton(wxButton& button, wxEvent& event)
  { dcDialogButtonPressed=wxOK; ((wxDialogBox*)button.GetParent())->Show(FALSE); }
void dcDialogCancelButton(wxButton& button, wxEvent& event)
  { dcDialogButtonPressed=wxCANCEL; ((wxDialogBox*)button.GetParent())->Show(FALSE); }
void dcDialogYesButton(wxButton& button, wxEvent& event)
  { dcDialogButtonPressed=wxYES; ((wxDialogBox*)button.GetParent())->Show(FALSE); }

char *wxGetTextWithHistory(char *message, char *caption, char *default_value,
                        wxFrame *parent, int x, int y)
{
  if (x < 0) x = wxDIALOG_DEFAULT_X;
  if (y < 0) y = wxDIALOG_DEFAULT_Y;

  wxDialogBox dialog(parent, caption, TRUE, x, y, 500, 160);
  dialog.SetClientSize(320, 160);

  (void)new wxMessage(&dialog, message);
  dialog.NewLine();

  wxText text(&dialog, NULL, NULL, default_value, -1, -1, 320);
  dialog.NewLine();

  wxButton *ok = new wxButton(&dialog, (wxFunction)&dcDialogOkButton, "OK");
  (void)new wxButton(&dialog, (wxFunction)&dcDialogCancelButton, "Cancel");
  (void)new wxButton(&dialog, (wxFunction)&dcDialogYesButton, "History");

  dcDialogButtonPressed = wxCANCEL;

  ok->SetDefault();
  text.SetFocus();

  dialog.Fit();
  dialog.Centre();
  dialog.Show(TRUE);

  if (dcDialogButtonPressed == wxOK || dcDialogButtonPressed == wxYES)
    return text.GetValue();
  else
    return NULL;
}

void dcwindow::suspend() { if (!inited) init(); frame->Iconize(TRUE); }

void dcwindow::select() 
{ 
if (!inited) init(); 
frame->Iconize(FALSE); 
#ifdef wx_motif
if (!resized)
  {
  int ww,hh;
//  frame->GetSize(&ww,&hh);
  frame->OnSize((c2-c1)*xscale,(r2-r1+2)*yscale*1.3);
  // this is supposed to get the scrollbars in the right place
  // but it doesn't work! I need to think of something else
  resized=1;
  }
#endif
}

int dcwindow::close() { if (!inited) init(); frame->Iconize(TRUE); return 1; }
void dcwindow::putch(char c) { select(); textwin->operator<<(c); }
void dcwindow::print(char *s) 
{ 
select(); 
textwin->WriteText(s); 
#if 0
wxYield(); 
#endif
}
void dcwindow::puts(char *s) { print(s); putch('\n'); }

char *dcwindow::fgets(char *s,int len,FILE *fp)
{
return ::fgets(s,len,fp);
}

dcwindow::~dcwindow() 
{ 
if (old_lines) delete old_lines;
#if 0
if (inited && parentptr==NULL) delete frame; 
// theory here is that parent will delete the frame
#endif
}

void dcwindow::set_history(int sz)
{
if (!old_lines) old_lines=new dczGDListchar(1);
history_size=sz;
}

dcwindow::dcwindow(wxFrame **p,char *t)
{
old_lines=0;
history_size=history_used=0;
inited=0;
#ifdef wx_motif
resized=0;
#endif
strcpy(title_str,t);
r1=c1=r2=c2=-1;
parentptr=p?p:(wxFrame **)&mainframeptr;
}

dcwindow::dcwindow(wxFrame **p,char *t,int rr1,int cc1,int rr2,int cc2) 
{
old_lines=0;
history_size=history_used=0;
inited=0;
#ifdef wx_motif
resized=0;
#endif
strcpy(title_str,t);
r1=rr1; c1=cc1; r2=rr2; c2=cc2;
parentptr=p?p:(wxFrame**)&mainframeptr;
}

void set_scale(wxWindow *w)
{
#if 1
#if defined wx_motif && !defined SCALEBUGFIXED
xscale=7;
yscale=16;
#else
xscale=w->GetCharWidth();
yscale=w->GetCharHeight();
#endif
#else
w->GetTextExtent("gggggggggg",&xscale,&yscale);
#ifdef wx_msw
    xscale/=11;
#else
    xscale/=9;
#endif
#endif
}

int dcwindow::init()
{
if (r1<0) frame=new wxFrame(*parentptr,title_str);
else
  {
  if (xscale==0.0) set_scale((wxWindow *)(*parentptr));
  frame=new wxFrame(*parentptr,title_str,int(c1*xscale),int(r1*yscale),int((c2-c1)*xscale),int((r2-r1+2)*yscale*1.3));
  }
textwin=new wxTextWindow(frame);
inited=1;
return 1;
}

int dcwindow::gets(char* s,int l,char *pr)
{
if (!inited) init();
char *t;
t=s;
if (pr) clear();
if (!history_size || !history_used)
  t=wxGetTextFromUser(pr?pr:"","Input text",t,*parentptr,int(c1*xscale),int(r1*yscale)); 
else
  {
  char local_buff[200];
  while ((t=wxGetTextWithHistory(pr?pr:"","Input text",t,*parentptr,
                         int(c1*xscale),int(r1*yscale))),
    dcDialogButtonPressed==wxYES)
      {
      strcpy(local_buff,t);
      t=local_buff;
      winmenu h(r1+1,c1+1,0,0);
      int c;
      if ((c=h.choose(*old_lines))!=0)
        {
        old_lines->start();
        while (--c>0) old_lines->operator++();
        t=old_lines->operator()();
        }
      }
  }
if (t!=NULL && strlen(t)>0 && strlen(t)<history_size)
  {
        {
        char *temp;
        while (history_size<history_used+strlen(t) && old_lines->size())
          {
          old_lines->end(); 
          temp=old_lines->linkout(); 
          history_used-=strlen(temp);
          delete temp;
          }
        if ((temp=new char[strlen(t)+1])!=0)
          {
          old_lines->start();
          strcpy(temp,t);
          history_used+=strlen(temp);
          old_lines->linkin(temp);
          old_lines->start();
          temp=old_lines->linkout();
          old_lines->linkin(temp);
          }
        }
  }
if (l==-1) l=DEFAULT_MAX_INPUT_LENGTH;
if (t!=NULL) 
  {
  strncpy(s,t,l-1);
  s[l-1]='\0';
  }
return t!=NULL;
}

int dcwindow::printf(char *format,...)
{
    va_list arg_ptr;
    va_start(arg_ptr,format);
    vprintf(format,arg_ptr);
return 1;
}

int dcwindow::vprintf(char *format,va_list arg_ptr)
{
char printbuf[2001];
vsprintf(printbuf,format,arg_ptr);
print(printbuf);
return 1;
}

